@@ -1,10 +1,11 @@
 
 'VisiData: a curses interface for exploring and arranging tabular data'
 
-from .vdtui import __version__
+from .vdtui import __version__, __version_info__
 from .vdtui import *
 from .Path import *
-from .async import *
+from .asyncthread import *
+from .diskcache import *
 from .zscroll import *
 from .aggregators import *
 from .data import *
@@ -21,6 +22,10 @@ from .freeze import *
 from .regex import *
 from .canvas import *
 from .graph import *
+from .menu import *
+from .motd import *
+from .transpose import *
+from .diff import *
 
 from .loaders.csv import *
 from .loaders.json import *
@@ -34,5 +39,13 @@ from .loaders.shp import *
 from .loaders.mbtiles import *
 from .loaders.http import *
 from .loaders.html import *
+from .loaders.markdown import *
+from .loaders.png import *
+from .loaders.ttf import *
+from .loaders.sas import *
+from .loaders.spss import *
+from .loaders.stata import *
+
+from math import *
 
 addGlobals(globals())
