@@ -7,9 +7,10 @@ from visidata import *
 option('show_graph_labels', True, 'show axes and legend on graph')
 option('plot_colors', 'green red yellow cyan magenta white 38 136 168', 'list of distinct colors to use for plotting distinct objects')
 option('disp_pixel_random', False, 'randomly choose attr from set of pixels instead of most common')
-option('zoom_incr', 2.0, 'amount to multiply current zoomlevel by when zooming')
+option('zoom_incr', 2.0, 'amount to multiply current zoomlevel when zooming')
 option('color_graph_hidden', '238 blue', 'color of legend for hidden attribute')
 
+
 class Point:
     def __init__(self, x, y):
         self.x = x
@@ -105,13 +106,13 @@ def iterline(x1, y1, x2, y2):
     xdir = 1 if x1 <= x2 else -1
     ydir = 1 if y1 <= y2 else -1
 
-    r = max(xdiff, ydiff)
+    r = math.ceil(max(xdiff, ydiff))
     if r == 0:  # point, not line
         yield x1, y1
     else:
-        x, y = x1, y1
+        x, y = math.floor(x1), math.floor(y1)
         i = 0
-        while i <= r:
+        while i < r:
             x += xdir * xdiff / r
             y += ydir * ydiff / r
 
@@ -126,9 +127,10 @@ def anySelected(vs, rows):
 
 #  - width/height are exactly equal to the number of pixels displayable, and can change at any time.
 #  - needs to refresh from source on resize
-class Plotter(Sheet):
+class Plotter(BaseSheet):
     'pixel-addressable display of entire terminal with (x,y) integer pixel coordinates'
     columns=[Column('')]  # to eliminate errors outside of draw()
+    rowtype='pixels'
     commands=[
         Command('^L', 'refresh()', 'redraw all pixels on canvas'),
         Command('v', 'options.show_graph_labels = not options.show_graph_labels', 'toggle show_graph_labels'),
@@ -136,11 +138,14 @@ class Plotter(Sheet):
     ]
     def __init__(self, name, **kwargs):
         super().__init__(name, **kwargs)
-        self.labels = []  # (x, y, text, attr)
+        self.labels = []  # (x, y, text, attr, row)
         self.hiddenAttrs = set()
         self.needsRefresh = False
         self.resetCanvasDimensions()
 
+    def __len__(self):
+        return (self.plotwidth* self.plotheight)
+
     def resetCanvasDimensions(self):
         'sets total available canvas dimensions'
         self.plotwidth = vd().windowWidth*2
@@ -154,10 +159,10 @@ class Plotter(Sheet):
 
     def plotline(self, x1, y1, x2, y2, attr, row=None):
         for x, y in iterline(x1, y1, x2, y2):
-            self.plotpixel(round(x), round(y), attr, row)
+            self.plotpixel(math.ceil(x), math.ceil(y), attr, row)
 
-    def plotlabel(self, x, y, text, attr):
-        self.labels.append((x, y, text, attr))
+    def plotlabel(self, x, y, text, attr, row=None):
+        self.labels.append((x, y, text, attr, row))
 
     def plotlegend(self, i, txt, attr):
         self.plotlabel(self.plotwidth-30, i*4, txt, attr)
@@ -174,17 +179,17 @@ class Plotter(Sheet):
     def getPixelAttrRandom(self, x, y):
         'weighted-random choice of attr at this pixel.'
         c = list(attr for attr, rows in self.pixels[y][x].items()
-                         for r in rows if attr not in self.hiddenAttrs)
+                         for r in rows if attr and attr not in self.hiddenAttrs)
         return random.choice(c) if c else 0
 
     def getPixelAttrMost(self, x, y):
         'most common attr at this pixel.'
         r = self.pixels[y][x]
-        c = sorted((len(rows), attr, rows) for attr, rows in r.items() if attr not in self.hiddenAttrs)
+        c = sorted((len(rows), attr, rows) for attr, rows in r.items() if attr and attr not in self.hiddenAttrs)
         if not c:
             return 0
         _, attr, rows = c[-1]
-        if anySelected(self.source, rows):
+        if isinstance(self.source, BaseSheet) and anySelected(self.source, rows):
             attr, _ = colors.update(attr, 8, 'bold', 10)
         return attr
 
@@ -247,9 +252,43 @@ class Plotter(Sheet):
                     if attr:
                         scr.addstr(char_y, char_x, chr(0x2800+braille_num), attr)
 
+        def _mark_overlap_text(labels, textobj):
+            def _overlaps(a, b):
+                a_x1, _, a_txt, _, _ = a
+                b_x1, _, b_txt, _, _ = b
+                a_x2 = a_x1 + len(a_txt)
+                b_x2 = b_x1 + len(b_txt)
+                if a_x1 < b_x1 < a_x2 or a_x1 < b_x2 < a_x2 or \
+                   b_x1 < a_x1 < b_x2 or b_x1 < a_x2 < b_x2:
+                   return True
+                else:
+                   return False
+
+            label_fldraw = [textobj, True]
+            labels.append(label_fldraw)
+            for o in labels:
+                if _overlaps(o[0], textobj):
+                    o[1] = False
+                    label_fldraw[1] = False
+
         if options.show_graph_labels:
-            for pix_x, pix_y, txt, attr in self.labels:
-                clipdraw(scr, int(pix_y/4), int(pix_x/2), txt, attr, len(txt))
+            labels_by_line = defaultdict(list) # y -> text labels
+
+            for pix_x, pix_y, txt, attr, row in self.labels:
+                if attr in self.hiddenAttrs:
+                    continue
+                if row is not None:
+                    pix_x -= len(txt)/2*2
+                char_y = int(pix_y/4)
+                char_x = int(pix_x/2)
+                o = (char_x, char_y, txt, attr, row)
+                _mark_overlap_text(labels_by_line[char_y], o)
+
+            for line in labels_by_line.values():
+                for o, fldraw in line:
+                    if fldraw:
+                        char_x, char_y, txt, attr, row = o
+                        clipdraw(scr, char_y, char_x, txt, attr, len(txt))
 
 
 # - has a cursor, of arbitrary position and width/height (not restricted to current zoom)
@@ -260,13 +299,17 @@ class Canvas(Plotter):
     leftMarginPixels = 10*2
     rightMarginPixels = 6*2
     topMarginPixels = 0
-    bottomMarginPixels = 2*4  # reserve bottom line for x axis
+    bottomMarginPixels = 1*4  # reserve bottom line for x axis
 
     commands = Plotter.commands + [
-        Command('move-left', 'sheet.cursorBox.xmin -= cursorBox.w', ''),
-        Command('move-right', 'sheet.cursorBox.xmin += cursorBox.w', ''),
-        Command('move-down', 'sheet.cursorBox.ymin += cursorBox.h', ''),
-        Command('move-up', 'sheet.cursorBox.ymin -= cursorBox.h', ''),
+        Command('view-go-left', 'sheet.cursorBox.xmin -= cursorBox.w', ''),
+        Command('view-go-right', 'sheet.cursorBox.xmin += cursorBox.w', ''),
+        Command('view-go-up', 'sheet.cursorBox.ymin += cursorBox.h', ''),
+        Command('view-go-down', 'sheet.cursorBox.ymin -= cursorBox.h', ''),
+        Command('view-go-far-left', 'sheet.cursorBox.xmin = visibleBox.xmin', ''),
+        Command('view-go-far-right', 'sheet.cursorBox.xmin = visibleBox.xmax-cursorBox.w', ''),
+        Command('view-go-far-up', 'sheet.cursorBox.ymin = cursorBox.ymin', ''),
+        Command('view-go-far-down', 'sheet.cursorBox.ymin = cursorBox.ymax-cursorBox.h', ''),
 
         Command('zh', 'sheet.cursorBox.xmin -= canvasCharWidth', ''),
         Command('zl', 'sheet.cursorBox.xmin += canvasCharWidth', ''),
@@ -303,13 +346,15 @@ class Canvas(Plotter):
         Command('s', 'source.select(list(rowsWithin(plotterCursorBox)))', 'select rows on source sheet contained within canvas cursor'),
         Command('t', 'source.toggle(list(rowsWithin(plotterCursorBox)))', 'toggle selection of rows on source sheet contained within canvas cursor'),
         Command('u', 'source.unselect(list(rowsWithin(plotterCursorBox)))', 'unselect rows on source sheet contained within canvas cursor'),
-        Command(ENTER, 'vs=copy(source); vs.rows=list(rowsWithin(plotterCursorBox)); vd.push(vs)', 'Open sheet of source rows contained within canvas cursor'),
+        Command(ENTER, 'vs=copy(source); vs.rows=list(rowsWithin(plotterCursorBox)); vd.push(vs)', 'open sheet of source rows contained within canvas cursor'),
+        Command('d', 'source.delete(list(rowsWithin(plotterCursorBox))); reload()', 'delete rows on source sheet contained within canvas cursor'),
 
 
-        Command('gs', 'source.select(list(rowsWithin(plotterVisibleBox)))', 'select rows visible on screen'),
-        Command('gt', 'source.toggle(list(rowsWithin(plotterVisibleBox)))', 'toggle selection of rows visible on screen'),
-        Command('gu', 'source.unselect(list(rowsWithin(plotterVisibleBox)))', 'unselect rows visible on screen'),
+        Command('gs', 'source.select(list(rowsWithin(plotterVisibleBox)))', 'select rows on source sheet visible on screen'),
+        Command('gt', 'source.toggle(list(rowsWithin(plotterVisibleBox)))', 'toggle selection of rows on source sheet visible on screen'),
+        Command('gu', 'source.unselect(list(rowsWithin(plotterVisibleBox)))', 'unselect rows on source sheet visible on screen'),
         Command('g'+ENTER, 'vs=copy(source); vs.rows=list(rowsWithin(plotterVisibleBox)); vd.push(vs)', 'open sheet of source rows visible on screen'),
+        Command('gd', 'source.delete(list(rowsWithin(plotterVisibleBox))); reload()', 'delete rows on source sheet visible on screen'),
     ]
 
     def __init__(self, name, source=None, **kwargs):
@@ -337,7 +382,7 @@ class Canvas(Plotter):
         self.polylines.clear()
         self.legends.clear()
         self.plotAttrs.clear()
-        self.unusedAttrs = list(colors[colorname.translate(str.maketrans('_', ' '))] for colorname in options.plot_colors.split())[::-1]
+        self.unusedAttrs = list(colors[colorname.translate(str.maketrans('_', ' '))] for colorname in options.plot_colors.split())
 
     def plotColor(self, k):
         attr = self.plotAttrs.get(k, None)
@@ -357,7 +402,7 @@ class Canvas(Plotter):
     def resetCanvasDimensions(self):
         super().resetCanvasDimensions()
         self.plotviewBox = BoundingBox(self.leftMarginPixels, self.topMarginPixels,
-                                       self.plotwidth-self.rightMarginPixels, self.plotheight-self.bottomMarginPixels)
+                                       self.plotwidth-self.rightMarginPixels, self.plotheight-self.bottomMarginPixels-1)
 
     @property
     def statusLine(self):
@@ -414,6 +459,71 @@ class Canvas(Plotter):
         'adds lines for (x,y) vertexes of a polygon'
         self.polylines.append((vertexes + [vertexes[0]], attr, row))
 
+    def qcurve(self, vertexes, attr, row=None):
+        'quadratic curve from vertexes[0] to vertexes[2] with control point at vertexes[1]'
+        assert len(vertexes) == 3, len(vertexes)
+        x1, y1 = vertexes[0]
+        x2, y2 = vertexes[1]
+        x3, y3 = vertexes[2]
+
+        self.point(x1, y1, attr, row)
+        self._recursive_bezier(x1, y1, x2, y2, x3, y3, attr, row)
+        self.point(x3, y3, attr, row)
+
+    def _recursive_bezier(self, x1, y1, x2, y2, x3, y3, attr, row, level=0):
+        'from http://www.antigrain.com/research/adaptive_bezier/'
+        m_approximation_scale = 10.0
+        m_distance_tolerance = (0.5 / m_approximation_scale) ** 2
+        m_angle_tolerance = 1 * 2*math.pi/360  # 15 degrees in rads
+        curve_angle_tolerance_epsilon = 0.01
+        curve_recursion_limit = 32
+        curve_collinearity_epsilon = 1e-30
+
+        if level > curve_recursion_limit:
+            return
+
+        # Calculate all the mid-points of the line segments
+
+        x12   = (x1 + x2) / 2
+        y12   = (y1 + y2) / 2
+        x23   = (x2 + x3) / 2
+        y23   = (y2 + y3) / 2
+        x123  = (x12 + x23) / 2
+        y123  = (y12 + y23) / 2
+
+        dx = x3-x1
+        dy = y3-y1
+        d = abs(((x2 - x3) * dy - (y2 - y3) * dx))
+
+        if d > curve_collinearity_epsilon:
+            # Regular care
+            if d*d <= m_distance_tolerance * (dx*dx + dy*dy):
+                # If the curvature doesn't exceed the distance_tolerance value, we tend to finish subdivisions.
+                if m_angle_tolerance < curve_angle_tolerance_epsilon:
+                    self.point(x123, y123, attr, row)
+                    return
+
+                # Angle & Cusp Condition
+                da = abs(math.atan2(y3 - y2, x3 - x2) - math.atan2(y2 - y1, x2 - x1))
+                if da >= math.pi:
+                    da = 2*math.pi - da
+
+                if da < m_angle_tolerance:
+                    # Finally we can stop the recursion
+                    self.point(x123, y123, attr, row)
+                    return
+        else:
+            # Collinear case
+            dx = x123 - (x1 + x3) / 2
+            dy = y123 - (y1 + y3) / 2
+            if dx*dx + dy*dy <= m_distance_tolerance:
+                self.point(x123, y123, attr, row)
+                return
+
+        # Continue subdivision
+        self._recursive_bezier(x1, y1, x12, y12, x123, y123, attr, row, level + 1)
+        self._recursive_bezier(x123, y123, x23, y23, x3, y3, attr, row, level + 1)
+
     def label(self, x, y, text, attr, row=None):
         self.gridlabels.append((x, y, text, attr, row))
 
@@ -516,7 +626,7 @@ class Canvas(Plotter):
         self.resetCanvasDimensions()
         self.render_async()
 
-    @async
+    @asyncthread
     def render_async(self):
         'plots points and lines and text onto the Plotter'
 
