@@ -8,25 +8,28 @@ option('replay_wait', 0.0, 'time to wait between replayed commands, in seconds')
 option('disp_replay_play', '▶', 'status indicator for active replay')
 option('disp_replay_pause', '‖', 'status indicator for paused replay')
 option('replay_movement', False, 'insert movements during replay')
+option('visidata_dir', '~/.visidata/', 'directory to load and store macros')
 
-globalCommand('D', 'vd.push(vd.cmdlog)', 'open CommandLog', 'open-cmdlog')
-globalCommand('^D', 'saveSheet(vd.cmdlog, inputFilename("save to: ", value=fnSuffix("cmdlog-{0}.vd") or "cmdlog.vd"))', 'save CommandLog to new .vd file', 'save-cmdlog')
-globalCommand('^U', 'CommandLog.togglePause()', 'pause/resume replay', 'toggle-replay')
-globalCommand('^I', '(CommandLog.currentReplay or error("no replay to advance")).advance()', 'execute next row in replaying sheet', 'step-replay')
-globalCommand('^K', '(CommandLog.currentReplay or error("no replay to cancel")).cancel()', 'cancel current replay', 'cancel-replay')
-
-#globalCommand('KEY_BACKSPACE', 'vd.cmdlog.undo()', 'remove last action on commandlog and replay')
+globalCommand('gD', 'p=Path(options.visidata_dir); vd.push(DirSheet(p.name, source=p))', 'open .visidata directory', 'meta-visidata-dir')
+globalCommand('D', 'vd.push(vd.cmdlog)', 'open CommandLog', 'meta-cmdlog')
+globalCommand('^D', 'saveSheets(inputFilename("save to: ", value=fnSuffix("cmdlog-{0}.vd") or "cmdlog.vd"), vd.cmdlog)', 'save CommandLog to new .vd file', 'meta-cmdlog-save')
+globalCommand('^U', 'CommandLog.togglePause()', 'pause/resume replay', 'meta-replay-toggle')
+globalCommand('^I', '(CommandLog.currentReplay or error("no replay to advance")).advance()', 'execute next row in replaying sheet', 'meta-replay-step')
+globalCommand('^K', '(CommandLog.currentReplay or error("no replay to cancel")).cancel()', 'cancel current replay', 'meta-replay-cancel')
 
+globalCommand('Q', 'vd.cmdlog.removeSheet(vd.sheets.pop(0))', 'quit current sheet and remove it from the cmdlog', 'sheet-quit-remove')
 
 globalCommand('status', 'status(input("status: ", display=False))', 'show given status message')
 
 # not necessary to log movements and scrollers
-nonLogKeys = 'KEY_DOWN KEY_UP KEY_NPAGE KEY_PPAGE kDOWN kUP j k gj gk ^F ^B r < > { } / ? n N gg G g/ g?'.split()
+nonLogKeys = 'KEY_DOWN KEY_UP KEY_NPAGE KEY_PPAGE j k gj gk ^F ^B r < > { } / ? n N gg G g/ g?'.split()
 nonLogKeys += 'KEY_LEFT KEY_RIGHT h l gh gl c'.split()
 nonLogKeys += 'zk zj zt zz zb zh zl zKEY_LEFT zKEY_RIGHT'.split()
-nonLogKeys += '^L ^C ^U ^K ^I ^D KEY_RESIZE'.split()
+nonLogKeys += '^L ^C ^U ^K ^I ^D ^G KEY_RESIZE KEY_F(1) z? KEY_BACKSPACE'.split()
+nonLogKeys += [' ']
 
 option('rowkey_prefix', 'キ', 'string prefix for rowkey in the cmdlog')
+option('cmdlog_histfile', '', 'file to autorecord each cmdlog action to')
 
 def itemsetter(i):
     def g(obj, v):
@@ -51,7 +54,7 @@ def namedlist(objname, fieldnames):
 CommandLogRow = namedlist('CommandLogRow', 'sheet col row keystrokes input comment'.split())
 
 def fnSuffix(template):
-    for i in range(1, 10):
+    for i in range(1, 1000):
         fn = template.format(i)
         if not Path(fn).exists():
             return fn
@@ -69,24 +72,34 @@ def getRowIdxByKey(sheet, k):
         if keystr(sheet.rowkey(r)) == k:
             return i
 
-def loggable(keystrokes):
+def isLoggableSheet(sheet):
+    return sheet is not vd().cmdlog and not isinstance(sheet, (OptionsSheet, ErrorSheet))
+
+def isLoggableCommand(keystrokes):
     if keystrokes in nonLogKeys:
         return False
     if keystrokes.startswith('move-'):
         return False
+    if keystrokes.startswith('BUTTON'):  # mouse click
+        return False
+    if keystrokes.startswith('REPORT'):  # scrollwheel/mouse position
+        return False
     return True
 
 def open_vd(p):
     return CommandLog(p.name, source=p)
 
-# rowdef: CommandLog
+save_vd = save_tsv
+
+# rowdef: CommandLogRow
 class CommandLog(Sheet):
     'Log of commands for current session.'
-    rowtype = 'commands'
+    rowtype = 'logged commands'
     commands = [
-        Command('x', 'sheet.replayOne(cursorRow); status("replayed one row")', 'replay command in current row'),
-        Command('gx', 'sheet.replay()', 'replay contents of entire CommandLog'),
-        Command('^C', 'sheet.cursorRowIndex = sheet.nRows', 'abort replay'),
+        Command('x', 'sheet.replayOne(cursorRow); status("replayed one row")', 'replay command in current row', 'replay-row'),
+        Command('gx', 'sheet.replay()', 'replay contents of entire CommandLog', 'replay-sheet'),
+        Command('^C', 'sheet.cursorRowIndex = sheet.nRows', 'abort replay', 'replay-abort'),
+        Command('z^S', 'sheet.saveMacro(selectedRows or error("no rows selected"), input("save macro for keystroke: "))', 'save macro', 'meta-macro-save'),
     ]
     columns = [ColumnAttr(x) for x in CommandLogRow._fields]
 
@@ -100,33 +113,29 @@ class CommandLog(Sheet):
         super().__init__(name, source=source, **kwargs)
         self.currentActiveRow = None
 
-        self.sheetmap = {}   # sheet.name -> vs
-
     def newRow(self):
         return CommandLogRow()
 
+    @asyncthread
     def reload(self):
         reload_tsv_sync(self, header=1)  # .vd files always have a header row, regardless of options
         self.rows = [CommandLogRow(r) for r in self.rows]
 
-    def undo(self):
-        'Delete last command, reload sources, and replay entire log.'
-        if len(self.rows) < 2:
-            error('no more to undo')
+    def removeSheet(self, vs):
+        'Remove all traces of sheets named vs.name from the cmdlog.'
+        self.rows = [r for r in self.rows if r.sheet != vs.name]
+        status('removed "%s" from cmdlog' % vs.name)
 
-        deletedRow = self.rows[-2]   # the command to undo
-        del self.rows[-2:]           # delete the previous command and the undo command
-
-        vd().sheets = [self]
-
-        self.sheetmap = {}
-        for r in self.rows:
-            self.replayOne(r)
-
-        status('undid "%s"' % deletedRow.keystrokes)
+    def saveMacro(self, rows, ks):
+        vs = copy(self)
+        vs.rows = self.selectedRows
+        macropath = Path(fnSuffix(options.visidata_dir+"macro-{0}.vd"))
+        save_vd(macropath, vs)
+        vd().macros[ks] = vs
+        append_tsv_row(vd().macrosheet, (ks, macropath.resolve()))
 
     def beforeExecHook(self, sheet, keystrokes, args=''):
-        if sheet is self:
+        if not isLoggableSheet(sheet):
             return  # don't record editlog commands
         if self.currentActiveRow:
             self.afterExecSheet(sheet, False, '')
@@ -153,26 +162,20 @@ class CommandLog(Sheet):
         if err:
             self.currentActiveRow[-1] += ' [%s]' % err
 
-        if sheet is not self:  # don't record jumps to cmdlog
+        if isLoggableSheet(sheet):  # don't record jumps to cmdlog or other internal sheets
             # remove user-aborted commands and simple movements
-            if not escaped and loggable(self.currentActiveRow.keystrokes):
+            if not escaped and isLoggableCommand(self.currentActiveRow.keystrokes):
                 self.addRow(self.currentActiveRow)
+                if options.cmdlog_histfile:
+                    if not getattr(vd(), 'sessionlog', None):
+                        vd().sessionlog = loadInternalSheet(CommandLog, Path(date().to_string(options.cmdlog_histfile)))
+                    append_tsv_row(vd().sessionlog, self.currentActiveRow)
 
         self.currentActiveRow = None
 
     def openHook(self, vs, src):
         self.addRow(CommandLogRow(['', '', '', 'o', src, 'open file']))
 
-    def getSheet(self, sheetname):
-        vs = self.sheetmap.get(sheetname)
-        if not vs:
-            matchingSheets = [x for x in vd().sheets if x.name == sheetname]
-            if not matchingSheets:
-                status(','.join(x.name for x in vd().sheets))
-                return None
-            vs = matchingSheets[0]
-        return vs
-
     @classmethod
     def togglePause(self):
         if not CommandLog.currentReplay:
@@ -196,8 +199,12 @@ class CommandLog(Sheet):
         if not r.sheet:
 #            assert not r.col and not r.row
             return self  # any old sheet should do, row/column don't matter
-        else:
-            vs = self.getSheet(r.sheet) or error('no sheets named %s' % r.sheet)
+
+        try:
+            sheetidx = int(r.sheet)
+            vs = vd().sheets[sheetidx]
+        except ValueError:
+            vs = vd().getSheet(r.sheet) or error('no sheet named %s' % r.sheet)
 
         if r.row:
             try:
@@ -207,7 +214,7 @@ class CommandLog(Sheet):
                 rowidx = getRowIdxByKey(vs, k)
 
             if rowidx is None:
-                error('no row %s' % r.row)
+                error('no "%s" row' % r.row)
 
             if options.replay_movement:
                 while vs.cursorRowIndex != rowidx:
@@ -224,7 +231,7 @@ class CommandLog(Sheet):
                 vcolidx = getColVisibleIdxByFullName(vs, r.col)
 
             if vcolidx is None:
-                error('no column %s' % r.col)
+                error('no "%s" column' % r.col)
 
             if options.replay_movement:
                 while vs.cursorVisibleColIndex != vcolidx:
@@ -259,7 +266,6 @@ class CommandLog(Sheet):
 
     def replay_sync(self, live=False):
         'Replay all commands in log.'
-        self.sheetmap.clear()
         self.cursorRowIndex = 0
         CommandLog.currentReplay = self
         with Progress(total=len(self.rows)) as prog:
@@ -269,8 +275,14 @@ class CommandLog(Sheet):
                     return
 
                 vd().statuses = []
-                if self.replayOne(self.cursorRow):
-                    CommandLog.currentReplay = None
+                try:
+                    if self.replayOne(self.cursorRow):
+                        self.cancel()
+                        return
+                except Exception as e:
+                    self.cancel()
+                    exceptionCaught(e)
+                    status('replay canceled')
                     return
 
                 self.cursorRowIndex += 1
@@ -283,7 +295,7 @@ class CommandLog(Sheet):
         status('replay complete')
         CommandLog.currentReplay = None
 
-    @async
+    @asyncthread
     def replay(self):
         'Inject commands into live execution with interface.'
         self.replay_sync(live=True)
@@ -306,6 +318,10 @@ class CommandLog(Sheet):
         x = options.disp_replay_pause if self.paused else options.disp_replay_play
         return ' │ %s %s/%s' % (x, self.cursorRowIndex, len(self.rows))
 
+    def setOption(self, optname, optval):
+        self.addRow(CommandLogRow(['options', '', options.rowkey_prefix + optname, 'set-row-input', str(optval), 'set option']))
+
+
 vd().cmdlog = CommandLog('cmdlog')
 vd().cmdlog.rows = []  # so it can be added to immediately
 
@@ -315,3 +331,15 @@ vd().addHook('preedit', vd().cmdlog.getLastArgs)
 vd().addHook('postedit', vd().cmdlog.setLastArgs)
 
 vd().addHook('rstatus', lambda sheet: CommandLog.currentReplay and (CommandLog.currentReplay.replayStatus, 'green'))
+vd().addHook('set_option', vd().cmdlog.setOption)
+
+def loadMacros():
+    macrospath = Path(os.path.join(options.visidata_dir, 'macros.tsv'))
+    macrosheet = loadInternalSheet(TsvSheet, macrospath, columns=(ColumnItem('keystrokes', 0), ColumnItem('filename', 1))) or error('error loading macros')
+
+    for ks, fn in macrosheet.rows:
+        vd().macros[ks] = loadInternalSheet(CommandLog, Path(fn))
+
+    return macrosheet
+
+vd().macrosheet = loadMacros()
