@@ -11,6 +11,10 @@ class HttpPath(PathFd):
 def openurl_http(p, filetype=None):
     import requests
     r = requests.get(p.url, stream=True)
+    if not filetype:
+        contenttype = r.headers['content-type']
+        if 'html' in contenttype:
+            filetype = 'html'
     return openSource(HttpPath(p.url, r.iter_lines(decode_unicode=True)), filetype=filetype)
 
 openurl_https = openurl_http
